@@ -1,5 +1,5 @@
 /* Common declarations for the tar program.
-   Copyright (C) 1988, 92, 93, 94, 96, 97 Free Software Foundation, Inc.
+   Copyright (C) 1988, 92, 93, 94, 96, 97, 1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -13,7 +13,7 @@
 
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
-   59 Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
 /* Declare the GNU tar archive format.  */
 #include "tar.h"
@@ -26,20 +26,6 @@
 #define PREFIX_FIELD_SIZE 155
 #define UNAME_FIELD_SIZE   32
 #define GNAME_FIELD_SIZE   32
-
-/* POSIX specified symbols currently unused are undefined here.  */
-#undef TSUID
-#undef TSGID
-#undef TSVTX
-#undef TUREAD
-#undef TUWRITE
-#undef TUEXEC
-#undef TGREAD
-#undef TGWRITE
-#undef TGEXEC
-#undef TOREAD
-#undef TOWRITE
-#undef TOEXEC
 
 /* Some various global definitions.  */
 
@@ -88,7 +74,11 @@ GLOBAL int exit_status;
 /* Information gleaned from the command line.  */
 
 #include "arith.h"
+#include "backupfile.h"
+#include "basename.h"
+#include "exclude.h"
 #include "modechange.h"
+#include "safe-read.h"
 
 /* Name of this program.  */
 GLOBAL const char *program_name;
@@ -121,7 +111,7 @@ GLOBAL char filename_terminator;
    not have _option in their name, even if their values is derived from
    option decoding, as these are especially important in tar.  */
 GLOBAL int blocking_factor;
-GLOBAL int record_size;
+GLOBAL size_t record_size;
 
 /* Boolean value.  */
 GLOBAL int absolute_names_option;
@@ -138,6 +128,9 @@ GLOBAL int atime_preserve_option;
 /* Boolean value.  */
 GLOBAL int backup_option;
 
+/* Type of backups being made.  */
+GLOBAL enum backup_type backup_type;
+
 /* Boolean value.  */
 GLOBAL int block_number_option;
 
@@ -150,8 +143,8 @@ GLOBAL const char *use_compress_program_option;
 /* Boolean value.  */
 GLOBAL int dereference_option;
 
-/* Boolean value.  */
-GLOBAL int exclude_option;
+/* Patterns that match file names to be excluded.  */
+GLOBAL struct exclude *excluded;
 
 /* Specified file containing names to work on.  */
 GLOBAL const char *files_from_option;
@@ -296,12 +289,12 @@ GLOBAL const char **archive_name_cursor;
 struct name
   {
     struct name *next;
-    short length;		/* cached strlen(name) */
+    size_t length;		/* cached strlen(name) */
     char found;			/* a matching file has been found */
     char firstch;		/* first char is literally matched */
     char regexp;		/* this name is a regexp, not literal */
     char *change_dir;		/* set with the -C option */
-    char *dir_contents;		/* for incremental_option */
+    const char *dir_contents;	/* for incremental_option */
     char fake;			/* dummy entry */
     char name[1];
   };
@@ -312,7 +305,7 @@ GLOBAL struct name *namelast;	/* points to last name in list */
 struct sp_array
   {
     off_t offset;
-    int numbytes;
+    size_t numbytes;
   };
 GLOBAL struct sp_array *sparsearray;
 
@@ -336,12 +329,12 @@ extern enum access_mode access_mode;
 
 extern FILE *stdlis;
 extern char *save_name;
-extern long save_sizeleft;
-extern long save_totsize;
+extern off_t save_sizeleft;
+extern off_t save_totsize;
 extern int write_archive_to_stdout;
 
-int available_space_after PARAMS ((union block *));
-long current_block_ordinal PARAMS ((void));
+size_t available_space_after PARAMS ((union block *));
+off_t current_block_ordinal PARAMS ((void));
 void close_archive PARAMS ((void));
 void closeout_volume_number PARAMS ((void));
 union block *find_next_block PARAMS ((void));
@@ -358,11 +351,30 @@ void set_next_block_after PARAMS ((union block *));
 /* Module create.c.  */
 
 void create_archive PARAMS ((void));
-void dump_file PARAMS ((char *, int, int));
+void dump_file PARAMS ((char *, dev_t, int));
 void finish_header PARAMS ((union block *));
-void to_oct PARAMS ((long, int, char *));
 void write_eot PARAMS ((void));
 
+#define GID_TO_OCT(val, where) gid_to_oct (val, where, sizeof (where))
+#define MAJOR_TO_OCT(val, where) major_to_oct (val, where, sizeof (where))
+#define MINOR_TO_OCT(val, where) minor_to_oct (val, where, sizeof (where))
+#define MODE_TO_OCT(val, where) mode_to_oct (val, where, sizeof (where))
+#define OFF_TO_OCT(val, where) off_to_oct (val, where, sizeof (where))
+#define SIZE_TO_OCT(val, where) size_to_oct (val, where, sizeof (where))
+#define TIME_TO_OCT(val, where) time_to_oct (val, where, sizeof (where))
+#define UID_TO_OCT(val, where) uid_to_oct (val, where, sizeof (where))
+#define UINTMAX_TO_OCT(val, where) uintmax_to_oct (val, where, sizeof (where))
+
+void gid_to_oct PARAMS ((gid_t, char *, size_t));
+void major_to_oct PARAMS ((major_t, char *, size_t));
+void minor_to_oct PARAMS ((minor_t, char *, size_t));
+void mode_to_oct PARAMS ((mode_t, char *, size_t));
+void off_to_oct PARAMS ((off_t, char *, size_t));
+void size_to_oct PARAMS ((size_t, char *, size_t));
+void time_to_oct PARAMS ((time_t, char *, size_t));
+void uid_to_oct PARAMS ((uid_t, char *, size_t));
+void uintmax_to_oct PARAMS ((uintmax_t, char *, size_t));
+
 /* Module diffarch.c.  */
 
 extern int now_verifying;
@@ -384,7 +396,7 @@ void delete_archive_members PARAMS ((void));
 /* Module incremen.c.  */
 
 void collect_and_sort_names PARAMS ((void));
-char *get_directory_contents PARAMS ((char *, int));
+char *get_directory_contents PARAMS ((char *, dev_t));
 void write_dir_file PARAMS ((void));
 void gnu_restore PARAMS ((int));
 void write_directory_file PARAMS ((void));
@@ -406,14 +418,37 @@ extern enum archive_format current_format;
 
 void decode_header PARAMS ((union block *, struct stat *,
 			    enum archive_format *, int));
-long from_oct PARAMS ((int, char *));
+#define STRINGIFY_BIGINT(i, b) \
+  stringify_uintmax_t_backwards ((uintmax_t) (i), (b) + sizeof (b))
+char *stringify_uintmax_t_backwards PARAMS ((uintmax_t, char *));
+
+#define GID_FROM_OCT(where) gid_from_oct (where, sizeof (where))
+#define MAJOR_FROM_OCT(where) major_from_oct (where, sizeof (where))
+#define MINOR_FROM_OCT(where) minor_from_oct (where, sizeof (where))
+#define MODE_FROM_OCT(where) mode_from_oct (where, sizeof (where))
+#define OFF_FROM_OCT(where) off_from_oct (where, sizeof (where))
+#define SIZE_FROM_OCT(where) size_from_oct (where, sizeof (where))
+#define TIME_FROM_OCT(where) time_from_oct (where, sizeof (where))
+#define UID_FROM_OCT(where) uid_from_oct (where, sizeof (where))
+#define UINTMAX_FROM_OCT(where) uintmax_from_oct (where, sizeof (where))
+
+gid_t gid_from_oct PARAMS ((const char *, size_t));
+major_t major_from_oct PARAMS ((const char *, size_t));
+minor_t minor_from_oct PARAMS ((const char *, size_t));
+mode_t mode_from_oct PARAMS ((const char *, size_t));
+off_t off_from_oct PARAMS ((const char *, size_t));
+size_t size_from_oct PARAMS ((const char *, size_t));
+time_t time_from_oct PARAMS ((const char *, size_t));
+uid_t uid_from_oct PARAMS ((const char *, size_t));
+uintmax_t uintmax_from_oct PARAMS ((const char *, size_t));
+
 void list_archive PARAMS ((void));
-void print_for_mkdir PARAMS ((char *, int, int));
+void print_for_mkdir PARAMS ((char *, int, mode_t));
 void print_header PARAMS ((void));
 void read_and PARAMS ((void (*do_) ()));
 enum read_header read_header PARAMS ((void));
 void skip_extended_headers PARAMS ((void));
-void skip_file PARAMS ((long));
+void skip_file PARAMS ((off_t));
 
 /* Module mangle.c.  */
 
@@ -455,10 +490,6 @@ char *name_from_list PARAMS ((void));
 void blank_name_list PARAMS ((void));
 char *new_name PARAMS ((const char *, const char *));
 
-void add_exclude PARAMS ((char *));
-void add_exclude_file PARAMS ((const char *));
-int check_exclude PARAMS ((const char *));
-
 /* Module tar.c.  */
 
 int confirm PARAMS ((const char *, const char *));
