@@ -1,5 +1,5 @@
 /* Create a tar archive.
-   Copyright (C) 1985, 92, 93, 94, 96, 97 Free Software Foundation, Inc.
+   Copyright 1985, 92, 93, 94, 96, 97, 1999 Free Software Foundation, Inc.
    Written by John Gilmore, on 1985-08-25.
 
    This program is free software; you can redistribute it and/or modify it
@@ -14,7 +14,7 @@
 
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
-   59 Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
 #include "system.h"
 
@@ -57,35 +57,141 @@ struct link *linklist = NULL;	/* points to first link in list */
 
 
 /*------------------------------------------------------------------------.
-| Converts long VALUE into a DIGS-digit field at WHERE, including a       |
-| trailing space and room for a NUL.  For example, 3 for DIGS 3 means one |
-| digit, a space, and room for a NUL.                                     |
+| Convert VALUE (with substitute SUBSTITUTE if VALUE is out of range)	  |
+| into a size-SIZE field at WHERE, including a				  |
+| trailing space.  For example, 3 for SIZE means two digits and a space.  |
 |                                                                         |
 | We assume the trailing NUL is already there and don't fill it in.  This |
 | fact is used by start_header and finish_header, so don't change it!     |
 `------------------------------------------------------------------------*/
 
-/* This should be equivalent to: sprintf (WHERE, "%*lo ", DIGS - 2, VALUE);
-   except that sprintf fills in the trailing NUL and we don't.  */
+/* Output VALUE in octal, using SUBSTITUTE if value won't fit.
+   Output to buffer WHERE with size SIZE.
+   TYPE is the kind of value being output (useful for diagnostics).
+   Prefer SIZE - 1 octal digits (with leading '0's), followed by '\0';
+   but if SIZE octal digits would fit, omit the '\0'.  */
 
-void
-to_oct (long value, int digs, char *where)
+static void
+to_oct (uintmax_t value, uintmax_t substitute, char *where, size_t size, const char *type)
 {
-  --digs;			/* Trailing null slot is left alone */
-  where[--digs] = ' ';		/* put in the space, though */
+  uintmax_t v = value;
+  size_t i = size;
+
+# define MAX_OCTAL_VAL_WITH_DIGITS(digits) \
+    ((digits) * 3 < sizeof (uintmax_t) * CHAR_BIT \
+     ? ((uintmax_t) 1 << ((digits) * 3)) - 1 \
+     : (uintmax_t) -1)
+
+  /* Output a trailing NUL unless the value is too large.  */
+  if (value <= MAX_OCTAL_VAL_WITH_DIGITS (size - 1))
+    where[--i] = '\0';
 
   /* Produce the digits -- at least one.  */
 
   do
     {
-      where[--digs] = '0' + (char) (value & 7);	/* one octal digit */
-      value >>= 3;
+      where[--i] = '0' + (int) (v & 7);	/* one octal digit */
+      v >>= 3;
     }
-  while (digs > 0 && value != 0);
+  while (i != 0 && v != 0);
 
-  /* Leading spaces, if necessary.  */
-  while (digs > 0)
-    where[--digs] = ' ';
+  /* Leading zeros, if necessary.  */
+  while (i != 0)
+    where[--i] = '0';
+
+  if (v != 0)
+    {
+      uintmax_t maxval = MAX_OCTAL_VAL_WITH_DIGITS (size);
+      char buf1[UINTMAX_STRSIZE_BOUND];
+      char buf2[UINTMAX_STRSIZE_BOUND];
+      char buf3[UINTMAX_STRSIZE_BOUND];
+      char *value_string = STRINGIFY_BIGINT (value, buf1);
+      char *maxval_string = STRINGIFY_BIGINT (maxval, buf2);
+      if (substitute)
+	{
+	  substitute &= maxval;
+	  WARN ((0, 0, _("%s value %s too large (max=%s); substituting %s"),
+		 type, value_string, maxval_string,
+		 STRINGIFY_BIGINT (substitute, buf3)));
+	  to_oct (substitute, (uintmax_t) 0, where, size, type);
+	}
+      else
+	ERROR ((0, 0, _("%s value %s too large (max=%s)"),
+		type, value_string, maxval_string));
+    }
+}
+#ifndef GID_NOBODY
+#define GID_NOBODY 0
+#endif
+void
+gid_to_oct (gid_t v, char *p, size_t s)
+{
+  to_oct ((uintmax_t) v, (uintmax_t) GID_NOBODY, p, s, "gid_t");
+}
+void
+major_to_oct (major_t v, char *p, size_t s)
+{
+  to_oct ((uintmax_t) v, (uintmax_t) 0, p, s, "major_t");
+}
+void
+minor_to_oct (minor_t v, char *p, size_t s)
+{
+  to_oct ((uintmax_t) v, (uintmax_t) 0, p, s, "minor_t");
+}
+void
+mode_to_oct (mode_t v, char *p, size_t s)
+{
+  /* In the common case where the internal and external mode bits are the same,
+     propagate all unknown bits to the external mode.
+     This matches historical practice.
+     Otherwise, just copy the bits we know about.  */
+  uintmax_t u =
+    ((S_ISUID == TSUID && S_ISGID == TSGID && S_ISVTX == TSVTX
+      && S_IRUSR == TUREAD && S_IWUSR == TUWRITE && S_IXUSR == TUEXEC
+      && S_IRGRP == TGREAD && S_IWGRP == TGWRITE && S_IXGRP == TGEXEC
+      && S_IROTH == TOREAD && S_IWOTH == TOWRITE && S_IXOTH == TOEXEC)
+     ? v
+     : ((v & S_ISUID ? TSUID : 0)
+	| (v & S_ISGID ? TSGID : 0)
+	| (v & S_ISVTX ? TSVTX : 0)
+	| (v & S_IRUSR ? TUREAD : 0)
+	| (v & S_IWUSR ? TUWRITE : 0)
+	| (v & S_IXUSR ? TUEXEC : 0)
+	| (v & S_IRGRP ? TGREAD : 0)
+	| (v & S_IWGRP ? TGWRITE : 0)
+	| (v & S_IXGRP ? TGEXEC : 0)
+	| (v & S_IROTH ? TOREAD : 0)
+	| (v & S_IWOTH ? TOWRITE : 0)
+	| (v & S_IXOTH ? TOEXEC : 0)));
+  to_oct (u, (uintmax_t) 0, p, s, "mode_t");
+}
+void
+off_to_oct (off_t v, char *p, size_t s)
+{
+  to_oct ((uintmax_t) v, (uintmax_t) 0, p, s, "off_t");
+}
+void
+size_to_oct (size_t v, char *p, size_t s)
+{
+  to_oct ((uintmax_t) v, (uintmax_t) 0, p, s, "size_t");
+}
+void
+time_to_oct (time_t v, char *p, size_t s)
+{
+  to_oct ((uintmax_t) v, (uintmax_t) 0, p, s, "time_t");
+}
+#ifndef UID_NOBODY
+#define UID_NOBODY 0
+#endif
+void
+uid_to_oct (uid_t v, char *p, size_t s)
+{
+  to_oct ((uintmax_t) v, (uintmax_t) UID_NOBODY, p, s, "uid_t");
+}
+void
+uintmax_to_oct (uintmax_t v, char *p, size_t s)
+{
+  to_oct (v, (uintmax_t) 0, p, s, "uintmax_t");
 }
 
 /* Writing routines.  */
@@ -114,9 +220,9 @@ write_eot (void)
 
   if (pointer)
     {
-      int space = available_space_after (pointer);
+      size_t space = available_space_after (pointer);
 
-      memset (pointer->buffer, 0, (size_t) space);
+      memset (pointer->buffer, 0, space);
       set_next_block_after (pointer);
     }
 }
@@ -132,8 +238,8 @@ static union block *start_header PARAMS ((const char *, struct stat *));
 static void
 write_long (const char *p, char type)
 {
-  int size = strlen (p) + 1;
-  int bufsize;
+  size_t size = strlen (p) + 1;
+  size_t bufsize;
   union block *header;
   struct stat foo;
 
@@ -150,15 +256,15 @@ write_long (const char *p, char type)
 
   while (bufsize < size)
     {
-      memcpy (header->buffer, p, (size_t) bufsize);
+      memcpy (header->buffer, p, bufsize);
       p += bufsize;
       size -= bufsize;
       set_next_block_after (header + (bufsize - 1) / BLOCKSIZE);
       header = find_next_block ();
       bufsize = available_space_after (header);
     }
-  memcpy (header->buffer, p, (size_t) size);
-  memset (header->buffer + size, 0, (size_t) (bufsize - size));
+  memcpy (header->buffer, p, size);
+  memset (header->buffer + size, 0, bufsize - size);
   set_next_block_after (header + (size - 1) / BLOCKSIZE);
 }
 
@@ -248,20 +354,20 @@ Removing leading `/' from absolute path names in the archive")));
      acceptor for Paul's test.  */
 
   if (archive_format == V7_FORMAT)
-    to_oct ((long) st->st_mode & 07777, 8, header->header.mode);
+    MODE_TO_OCT (st->st_mode & MODE_ALL, header->header.mode);
   else
-    to_oct ((long) st->st_mode, 8, header->header.mode);
+    MODE_TO_OCT (st->st_mode, header->header.mode);
 
-  to_oct ((long) st->st_uid, 8, header->header.uid);
-  to_oct ((long) st->st_gid, 8, header->header.gid);
-  to_oct ((long) st->st_size, 1 + 12, header->header.size);
-  to_oct ((long) st->st_mtime, 1 + 12, header->header.mtime);
+  UID_TO_OCT (st->st_uid, header->header.uid);
+  GID_TO_OCT (st->st_gid, header->header.gid);
+  OFF_TO_OCT (st->st_size, header->header.size);
+  TIME_TO_OCT (st->st_mtime, header->header.mtime);
 
   if (incremental_option)
     if (archive_format == OLDGNU_FORMAT)
       {
-	to_oct ((long) st->st_atime, 1 + 12, header->oldgnu_header.atime);
-	to_oct ((long) st->st_ctime, 1 + 12, header->oldgnu_header.ctime);
+	TIME_TO_OCT (st->st_atime, header->oldgnu_header.atime);
+	TIME_TO_OCT (st->st_ctime, header->oldgnu_header.ctime);
       }
 
   header->header.typeflag = archive_format == V7_FORMAT ? AREGTYPE : REGTYPE;
@@ -305,29 +411,29 @@ Removing leading `/' from absolute path names in the archive")));
 void
 finish_header (union block *header)
 {
-  int i, sum;
+  size_t i;
+  int sum;
   char *p;
 
   memcpy (header->header.chksum, CHKBLANKS, sizeof (header->header.chksum));
 
   sum = 0;
   p = header->buffer;
-  for (i = sizeof (*header); --i >= 0; )
+  for (i = sizeof (*header); i-- != 0; )
     /* We can't use unsigned char here because of old compilers, e.g. V7.  */
     sum += 0xFF & *p++;
 
   /* Fill in the checksum field.  It's formatted differently from the
      other fields: it has [6] digits, a null, then a space -- rather than
-     digits, a space, then a null.  We use to_oct then write the null in
-     over to_oct's space.  The final space is already there, from
+     digits, then a null.  We use to_oct.
+     The final space is already there, from
      checksumming, and to_oct doesn't modify it.
 
      This is a fast way to do:
 
      sprintf(header->header.chksum, "%6o", sum);  */
 
-  to_oct ((long) sum, 8, header->header.chksum);
-  header->header.chksum[6] = '\0';	/* zap the space */
+  uintmax_to_oct ((uintmax_t) sum, header->header.chksum, 7);
 
   set_next_block_after (header);
 
@@ -390,7 +496,7 @@ init_sparsearray (void)
 `---*/
 
 static void
-find_new_file_size (int *filesize, int highest_index)
+find_new_file_size (off_t *filesize, int highest_index)
 {
   int counter;
 
@@ -417,11 +523,11 @@ find_new_file_size (int *filesize, int highest_index)
 static int
 deal_with_sparse (char *name, union block *header)
 {
-  long numbytes = 0;
-  long offset = 0;
+  size_t numbytes = 0;
+  off_t offset = 0;
   int file;
   int sparse_index = 0;
-  int count;
+  ssize_t count;
   char buffer[BLOCKSIZE];
 
   if (archive_format == OLDGNU_FORMAT)
@@ -434,7 +540,7 @@ deal_with_sparse (char *name, union block *header)
   init_sparsearray ();
   clear_buffer (buffer);
 
-  while (count = read (file, buffer, sizeof buffer), count != 0)
+  while (count = safe_read (file, buffer, sizeof buffer), count != 0)
     {
       /* Realloc the scratch area as necessary.  FIXME: should reallocate
 	 only at beginning of a new instance of non-zero data.  */
@@ -511,14 +617,12 @@ deal_with_sparse (char *name, union block *header)
 `---*/
 
 static int
-finish_sparse_file (int file, long *sizeleft, long fullsize, char *name)
+finish_sparse_file (int file, off_t *sizeleft, off_t fullsize, char *name)
 {
   union block *start;
-  int bufsize;
+  size_t bufsize;
   int sparse_index = 0;
-  int count;
-  long pos;
-  long nwritten = 0;
+  ssize_t count;
 
   while (*sizeleft > 0)
     {
@@ -528,12 +632,24 @@ finish_sparse_file (int file, long *sizeleft, long fullsize, char *name)
       if (!bufsize)
 	{
 	  /* We blew it, maybe.  */
+	  char buf1[UINTMAX_STRSIZE_BOUND];
+	  char buf2[UINTMAX_STRSIZE_BOUND];
 
-	  ERROR ((0, 0, _("Wrote %ld of %ld bytes to file %s"),
-		  fullsize - *sizeleft, fullsize, name));
+	  ERROR ((0, 0, _("Wrote %s of %s bytes to file %s"),
+		  STRINGIFY_BIGINT (fullsize - *sizeleft, buf1),
+		  STRINGIFY_BIGINT (fullsize, buf2),
+		  name));
+	  break;
+	}
+
+      if (lseek (file, sparsearray[sparse_index++].offset, SEEK_SET) < 0)
+	{
+	  char buf[UINTMAX_STRSIZE_BOUND];
+	  ERROR ((0, errno, _("lseek error at byte %s in file %s"),
+		  STRINGIFY_BIGINT (sparsearray[sparse_index - 1].offset, buf),
+		  name));
 	  break;
 	}
-      pos = lseek (file, sparsearray[sparse_index++].offset, 0);
 
       /* If the number of bytes to be written here exceeds the size of
 	 the temporary buffer, do it in steps.  */
@@ -543,8 +659,8 @@ finish_sparse_file (int file, long *sizeleft, long fullsize, char *name)
 #if 0
 	  if (amount_read)
 	    {
-	      count = read (file, start->buffer + amount_read,
-			    BLOCKSIZE - amount_read);
+	      count = safe_read (file, start->buffer + amount_read,
+				 BLOCKSIZE - amount_read);
 	      bufsize -= BLOCKSIZE - amount_read;
 	      amount_read = 0;
 	      set_next_block_after (start);
@@ -554,18 +670,19 @@ finish_sparse_file (int file, long *sizeleft, long fullsize, char *name)
 #endif
 	  /* Store the data.  */
 
-	  count = read (file, start->buffer, BLOCKSIZE);
+	  count = safe_read (file, start->buffer, BLOCKSIZE);
 	  if (count < 0)
 	    {
+	      char buf[UINTMAX_STRSIZE_BOUND];
 	      ERROR ((0, errno, _("\
-Read error at byte %ld, reading %d bytes, in file %s"),
-			 fullsize - *sizeleft, bufsize, name));
+Read error at byte %s, reading %lu bytes, in file %s"),
+		      STRINGIFY_BIGINT (fullsize - *sizeleft, buf),
+		      (unsigned long) bufsize, name));
 	      return 1;
 	    }
 	  bufsize -= count;
 	  *sizeleft -= count;
 	  set_next_block_after (start);
-	  nwritten += BLOCKSIZE;	/* FIXME: ??? */
 	  start = find_next_block ();
 	  memset (start->buffer, 0, BLOCKSIZE);
 	}
@@ -574,15 +691,18 @@ Read error at byte %ld, reading %d bytes, in file %s"),
 	char buffer[BLOCKSIZE];
 
 	clear_buffer (buffer);
-	count = read (file, buffer, (size_t) bufsize);
+	count = safe_read (file, buffer, bufsize);
 	memcpy (start->buffer, buffer, BLOCKSIZE);
       }
 
       if (count < 0)
 	{
+	  char buf[UINTMAX_STRSIZE_BOUND];
+	  
 	  ERROR ((0, errno,
-		  _("Read error at byte %ld, reading %d bytes, in file %s"),
-		  fullsize - *sizeleft, bufsize, name));
+		  _("Read error at byte %s, reading %lu bytes, in file %s"),
+		  STRINGIFY_BIGINT (fullsize - *sizeleft, buf),
+		  (unsigned long) bufsize, name));
 	  return 1;
 	}
 #if 0
@@ -593,8 +713,8 @@ Read error at byte %ld, reading %d bytes, in file %s"),
 	  if (count != bufsize)
 	    {
 	      ERROR ((0, 0,
-		      _("File %s shrunk by %d bytes, padding with zeros"),
-		      name, sizeleft));
+		      _("File %s shrunk, padding with zeros"),
+		      name));
 	      return 1;
 	    }
 	  start = find_next_block ();
@@ -602,14 +722,12 @@ Read error at byte %ld, reading %d bytes, in file %s"),
       else
 	amount_read += bufsize;
 #endif
-      nwritten += count;	/* FIXME: ??? */
       *sizeleft -= count;
       set_next_block_after (start);
 
     }
   free (sparsearray);
 #if 0
-  printf (_("Amount actually written is (I hope) %d.\n"), nwritten);
   set_next_block_after (start + (count - 1) / BLOCKSIZE);
 #endif
   return 0;
@@ -631,12 +749,13 @@ create_archive (void)
   if (incremental_option)
     {
       char *buffer = xmalloc (PATH_MAX);
-      char *q, *bufp;
+      const char *q;
+      char *bufp;
 
       collect_and_sort_names ();
 
       while (p = name_from_list (), p)
-	dump_file (p, -1, 1);
+	dump_file (p, (dev_t) -1, 1);
 
       blank_name_list ();
       while (p = name_from_list (), p)
@@ -652,7 +771,7 @@ create_archive (void)
 	      if (*q == 'Y')
 		{
 		  strcpy (bufp, q + 1);
-		  dump_file (buffer, -1, 1);
+		  dump_file (buffer, (dev_t) -1, 1);
 		}
 	    }
 	}
@@ -661,7 +780,7 @@ create_archive (void)
   else
     {
       while (p = name_next (1), p)
-	dump_file (p, -1, 1);
+	dump_file (p, (dev_t) -1, 1);
     }
 
   write_eot ();
@@ -683,13 +802,14 @@ create_archive (void)
    exit_status to failure, a clear diagnostic has been issued.  */
 
 void
-dump_file (char *p, int parent_device, int top_level)
+dump_file (char *p, dev_t parent_device, int top_level)
 {
   union block *header;
   char type;
   union block *exhdr;
   char save_typeflag;
   struct utimbuf restore_times;
+  off_t restore_size;
 
   /* FIXME: `header' and `upperbound' might be used uninitialized in this
      function.  Reported by Bruno Haible.  */
@@ -701,7 +821,7 @@ dump_file (char *p, int parent_device, int top_level)
      Otherwise, use lstat (which falls back to stat if no symbolic links).  */
 
   if (dereference_option != 0
-#ifdef STX_HIDDEN		/* AIX */
+#if STX_HIDDEN && !_LARGE_FILES /* AIX */
       ? statx (p, &current_stat, STATSIZE, STX_HIDDEN)
       : statx (p, &current_stat, STATSIZE, STX_HIDDEN | STX_LINK)
 #else
@@ -717,6 +837,7 @@ dump_file (char *p, int parent_device, int top_level)
 
   restore_times.actime = current_stat.st_atime;
   restore_times.modtime = current_stat.st_mtime;
+  restore_size = current_stat.st_size;
 
 #ifdef S_ISHIDDEN
   if (S_ISHIDDEN (current_stat.st_mode))
@@ -738,7 +859,7 @@ dump_file (char *p, int parent_device, int top_level)
       && current_stat.st_mtime < newer_mtime_option
       && (!after_date_option || current_stat.st_ctime < newer_ctime_option))
     {
-      if (parent_device == -1)
+      if (parent_device == (dev_t) -1)
 	WARN ((0, 0, _("%s: is unchanged; not dumped"), p));
       /* FIXME: recheck this return.  */
       return;
@@ -850,8 +971,9 @@ Removing leading `/' from absolute links")));
       )
     {
       int f;			/* file descriptor */
-      long bufsize, count;
-      long sizeleft;
+      size_t bufsize;
+      ssize_t count;
+      off_t sizeleft;
       union block *start;
       int header_moved;
       char isextended = 0;
@@ -873,10 +995,6 @@ Removing leading `/' from absolute links")));
 	     files not having more hole blocks than indirect blocks, the
 	     sparseness will go undetected.  */
 
-	  /* tar.h defines ST_NBLOCKS in term of 512 byte sectors, even
-	     for HP-UX's which count in 1024 byte units and AIX's which
-	     count in 4096 byte units.  So this should work...  */
-
 	  /* Bruno Haible sent me these statistics for Linux.  It seems
 	     that some filesystems count indirect blocks in st_blocks,
 	     while others do not seem to:
@@ -899,9 +1017,11 @@ Removing leading `/' from absolute links")));
 	     st_blocks, so `du' and `ls -s' give wrong results.  So, the
 	     --sparse option would not work on a minix filesystem.  */
 
-	  if (current_stat.st_size > ST_NBLOCKS (current_stat) * BLOCKSIZE)
+	  if (ST_NBLOCKS (current_stat)
+	      < (current_stat.st_size / ST_NBLOCKSIZE
+		 + (current_stat.st_size % ST_NBLOCKSIZE != 0)))
 	    {
-	      int filesize = current_stat.st_size;
+	      off_t filesize = current_stat.st_size;
 	      int counter;
 
 	      header = start_header (p, &current_stat);
@@ -930,8 +1050,8 @@ Removing leading `/' from absolute links")));
 		 <file>.  It might be kind of disconcerting if the
 		 shrunken file size was the one that showed up.  */
 
-	      to_oct ((long) current_stat.st_size, 1 + 12,
-		      header->oldgnu_header.realsize);
+	      OFF_TO_OCT (current_stat.st_size,
+			  header->oldgnu_header.realsize);
 
 	      /* This will be the new "size" of the file, i.e., the size
 		 of the file minus the blocks of holes that we're
@@ -939,17 +1059,17 @@ Removing leading `/' from absolute links")));
 
 	      find_new_file_size (&filesize, upperbound);
 	      current_stat.st_size = filesize;
-	      to_oct ((long) filesize, 1 + 12, header->header.size);
+	      OFF_TO_OCT (filesize, header->header.size);
 
 	      for (counter = 0; counter < SPARSES_IN_OLDGNU_HEADER; counter++)
 		{
 		  if (!sparsearray[counter].numbytes)
 		    break;
 
-		  to_oct (sparsearray[counter].offset, 1 + 12,
-			  header->oldgnu_header.sp[counter].offset);
-		  to_oct (sparsearray[counter].numbytes, 1 + 12,
-			  header->oldgnu_header.sp[counter].numbytes);
+		  OFF_TO_OCT (sparsearray[counter].offset,
+			      header->oldgnu_header.sp[counter].offset);
+		  SIZE_TO_OCT (sparsearray[counter].numbytes,
+			       header->oldgnu_header.sp[counter].numbytes);
 		}
 
 	    }
@@ -963,7 +1083,8 @@ Removing leading `/' from absolute links")));
 	 files when archive is meant for /dev/null.  */
 
       if (dev_null_output
-	  || (sizeleft == 0 && 0444 == (0444 & current_stat.st_mode)))
+	  || (sizeleft == 0
+	      && MODE_R == (MODE_R & current_stat.st_mode)))
 	f = -1;
       else
 	{
@@ -1025,12 +1146,10 @@ Removing leading `/' from absolute links")));
 	      if (counter + index_offset > upperbound)
 		break;
 
-	      to_oct ((long) sparsearray[counter + index_offset].numbytes,
-		      1 + 12,
-		      exhdr->sparse_header.sp[counter].numbytes);
-	      to_oct ((long) sparsearray[counter + index_offset].offset,
-		      1 + 12,
-		      exhdr->sparse_header.sp[counter].offset);
+	      SIZE_TO_OCT (sparsearray[counter + index_offset].numbytes,
+			   exhdr->sparse_header.sp[counter].numbytes);
+	      OFF_TO_OCT (sparsearray[counter + index_offset].offset,
+			  exhdr->sparse_header.sp[counter].offset);
 	    }
 	  set_next_block_after (exhdr);
 #if 0
@@ -1048,7 +1167,8 @@ Removing leading `/' from absolute links")));
 	}
       if (save_typeflag == GNUTYPE_SPARSE)
 	{
-	  if (finish_sparse_file (f, &sizeleft, current_stat.st_size, p))
+	  if (f < 0
+	      || finish_sparse_file (f, &sizeleft, current_stat.st_size, p))
 	    goto padit;
 	}
       else
@@ -1068,7 +1188,7 @@ Removing leading `/' from absolute links")));
 	      {
 		/* Last read -- zero out area beyond.  */
 
-		bufsize = (int) sizeleft;
+		bufsize = sizeleft;
 		count = bufsize % BLOCKSIZE;
 		if (count)
 		  memset (start->buffer + sizeleft, 0,
@@ -1077,12 +1197,15 @@ Removing leading `/' from absolute links")));
 	    if (f < 0)
 	      count = bufsize;
 	    else
-	      count = read (f, start->buffer, (size_t) bufsize);
+	      count = safe_read (f, start->buffer, bufsize);
 	    if (count < 0)
 	      {
+		char buf[UINTMAX_STRSIZE_BOUND];
 		ERROR ((0, errno, _("\
-Read error at byte %ld, reading %d bytes, in file %s"),
-			(long) (current_stat.st_size - sizeleft), bufsize, p));
+Read error at byte %s, reading %lu bytes, in file %s"),
+			STRINGIFY_BIGINT (current_stat.st_size - sizeleft,
+					  buf),
+			(unsigned long) bufsize, p));
 		goto padit;
 	      }
 	    sizeleft -= count;
@@ -1093,9 +1216,14 @@ Read error at byte %ld, reading %d bytes, in file %s"),
 
 	    if (count == bufsize)
 	      continue;
-	    ERROR ((0, 0, _("File %s shrunk by %d bytes, padding with zeros"),
-		    p, sizeleft));
-	    goto padit;		/* short read */
+	    else
+	      {
+		char buf[UINTMAX_STRSIZE_BOUND];
+		ERROR ((0, 0,
+			_("File %s shrunk by %s bytes, padding with zeros"),
+			p, STRINGIFY_BIGINT (sizeleft, buf)));
+		goto padit;		/* short read */
+	      }
 	  }
 
       if (multi_volume_option)
@@ -1103,7 +1231,14 @@ Read error at byte %ld, reading %d bytes, in file %s"),
 
       if (f >= 0)
 	{
-	  close (f);
+	  struct stat final_stat;
+	  if (fstat (f, &final_stat) != 0)
+	    ERROR ((0, errno, "%s: fstat", p));
+	  else if (final_stat.st_mtime != restore_times.modtime
+		   || final_stat.st_size != restore_size)
+	    ERROR ((0, errno, _("%s: file changed as we read it"), p));
+	  if (close (f) != 0)
+	    ERROR ((0, errno, _("%s: close"), p));
 	  if (atime_preserve_option)
 	    utime (p, &restore_times);
 	}
@@ -1181,9 +1316,9 @@ Read error at byte %ld, reading %d bytes, in file %s"),
       DIR *directory;
       struct dirent *entry;
       char *namebuf;
-      int buflen;
-      int len;
-      int our_device = current_stat.st_dev;
+      size_t buflen;
+      size_t len;
+      dev_t our_device = current_stat.st_dev;
 
       /* If this tar program is installed suid root, like for Amanda, the
 	 access might look like denied, while it is not really.
@@ -1204,8 +1339,8 @@ Read error at byte %ld, reading %d bytes, in file %s"),
 
       len = strlen (p);
       buflen = len + NAME_FIELD_SIZE;
-      namebuf = xmalloc ((size_t) (buflen + 1));
-      strncpy (namebuf, p, (size_t) buflen);
+      namebuf = xmalloc (buflen + 1);
+      strncpy (namebuf, p, buflen);
       while (len >= 1 && namebuf[len - 1] == '/')
 	len--;
       namebuf[len++] = '/';
@@ -1254,25 +1389,25 @@ Read error at byte %ld, reading %d bytes, in file %s"),
 
       if (incremental_option && gnu_list_name->dir_contents)
 	{
-	  int sizeleft;
-	  int totsize;
-	  int bufsize;
+	  off_t sizeleft;
+	  off_t totsize;
+	  size_t bufsize;
 	  union block *start;
-	  int count;
-	  char *buffer, *p_buffer;
+	  ssize_t count;
+	  const char *buffer, *p_buffer;
 
 	  buffer = gnu_list_name->dir_contents; /* FOO */
 	  totsize = 0;
 	  for (p_buffer = buffer; p_buffer && *p_buffer;)
 	    {
-	      int tmp;
+	      size_t tmp;
 
 	      tmp = strlen (p_buffer) + 1;
 	      totsize += tmp;
 	      p_buffer += tmp;
 	    }
 	  totsize++;
-	  to_oct ((long) totsize, 1 + 12, header->header.size);
+	  OFF_TO_OCT (totsize, header->header.size);
 	  finish_header (header);
 	  p_buffer = buffer;
 	  sizeleft = totsize;
@@ -1294,7 +1429,7 @@ Read error at byte %ld, reading %d bytes, in file %s"),
 		    memset (start->buffer + sizeleft, 0,
 			   (size_t) (BLOCKSIZE - count));
 		}
-	      memcpy (start->buffer, p_buffer, (size_t) bufsize);
+	      memcpy (start->buffer, p_buffer, bufsize);
 	      sizeleft -= bufsize;
 	      p_buffer += bufsize;
 	      set_next_block_after (start + (bufsize - 1) / BLOCKSIZE);
@@ -1343,15 +1478,16 @@ Read error at byte %ld, reading %d bytes, in file %s"),
 
       while (entry = readdir (directory), entry)
 	{
-	  /* Skip `.' and `..'.  */
+	  /* Skip `.', `..', and excluded file names.  */
 
-	  if (is_dot_or_dotdot (entry->d_name))
+	  if (is_dot_or_dotdot (entry->d_name)
+	      || excluded_filename (excluded, entry->d_name))
 	    continue;
 
 	  if ((int) NAMLEN (entry) + len >= buflen)
 	    {
 	      buflen = len + NAMLEN (entry);
-	      namebuf = (char *) xrealloc (namebuf, (size_t) (buflen + 1));
+	      namebuf = (char *) xrealloc (namebuf, buflen + 1);
 #if 0
 	      namebuf[len] = '\0';
 	      ERROR ((0, 0, _("File name %s%s too long"),
@@ -1360,8 +1496,6 @@ Read error at byte %ld, reading %d bytes, in file %s"),
 #endif
 	    }
 	  strcpy (namebuf + len, entry->d_name);
-	  if (exclude_option && check_exclude (namebuf))
-	    continue;
 	  dump_file (namebuf, our_device, 0);
 	}
 
@@ -1413,10 +1547,8 @@ Read error at byte %ld, reading %d bytes, in file %s"),
 #if defined(S_IFBLK) || defined(S_IFCHR)
   if (type != FIFOTYPE)
     {
-      to_oct ((long) major (current_stat.st_rdev), 8,
-	      header->header.devmajor);
-      to_oct ((long) minor (current_stat.st_rdev), 8,
-	      header->header.devminor);
+      MAJOR_TO_OCT (major (current_stat.st_rdev), header->header.devmajor);
+      MINOR_TO_OCT (minor (current_stat.st_rdev), header->header.devminor);
     }
 #endif
 
