@@ -13,24 +13,21 @@
 
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
-   59 Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
 #include "system.h"
 
 #include <time.h>
 time_t time ();
 
-#define ISDIGIT(Char) (ISASCII (Char) && isdigit (Char))
-#define ISSPACE(Char) (ISASCII (Char) && isspace (Char))
-
 #include "common.h"
 
 /* Variable sized generic character buffers.  */
 
 struct accumulator
 {
-  int allocated;
-  int length;
+  size_t allocated;
+  size_t length;
   char *pointer;
 };
 
@@ -80,15 +77,15 @@ delete_accumulator (struct accumulator *accumulator)
 
 static void
 add_to_accumulator (struct accumulator *accumulator,
-		    const char *data, int size)
+		    const char *data, size_t size)
 {
   if (accumulator->length + size > accumulator->allocated)
     {
       accumulator->allocated = accumulator->length + size + ACCUMULATOR_SLACK;
       accumulator->pointer = (char *)
-	xrealloc (accumulator->pointer, (size_t) accumulator->allocated);
+	xrealloc (accumulator->pointer, accumulator->allocated);
     }
-  memcpy (accumulator->pointer + accumulator->length, data, (size_t) size);
+  memcpy (accumulator->pointer + accumulator->length, data, size);
   accumulator->length += size;
 }
 
@@ -102,21 +99,29 @@ struct directory
   {
     struct directory *next;	/* next entry in list */
     const char *name;		/* path name of directory */
-    int device_number;		/* device number for directory */
-    int inode_number;		/* inode number for directory */
-    int allnew;
+    dev_t device_number;	/* device number for directory */
+    ino_t inode_number;		/* inode number for directory */
+    char allnew;
+    char nfs;
     const char *dir_text;
   };
 static struct directory *directory_list = NULL;
 
+#if HAVE_ST_FSTYPE_STRING
+  static const char nfs[] = "nfs";
+# define NFS_FILE_STAT(st) (strcmp ((st).st_fstype, nfs) == 0)
+#else
+# define ST_DEV_MSB(st) (~ (dev_t) 0 << (sizeof (st).st_dev * CHAR_BIT - 1))
+# define NFS_FILE_STAT(st) (((st).st_dev & ST_DEV_MSB (st)) != 0)
+#endif
+
 /*-------------------------------------------------------------------.
 | Create and link a new directory entry for directory NAME, having a |
 | DEVICE_NUMBER and a INODE_NUMBER, with some TEXT.		     |
 `-------------------------------------------------------------------*/
 
 static void
-note_directory (char *name, dev_t device_number, ino_t inode_number,
-		const char *text)
+note_directory (char *name, struct stat *st, const char *text)
 {
   struct directory *directory
     = (struct directory *) xmalloc (sizeof (struct directory));
@@ -124,11 +129,12 @@ note_directory (char *name, dev_t device_number, ino_t inode_number,
   directory->next = directory_list;
   directory_list = directory;
 
-  directory->device_number = device_number;
-  directory->inode_number = inode_number;
+  directory->device_number = st->st_dev;
+  directory->inode_number = st->st_ino;
   directory->name = xstrdup (name);
   directory->dir_text = text;
   directory->allnew = 0;
+  directory->nfs = NFS_FILE_STAT (*st);
 }
 
 /*------------------------------------------------------------------------.
@@ -166,7 +172,7 @@ compare_dirents (const voidstar first, const voidstar second)
 `---*/
 
 char *
-get_directory_contents (char *path, int device)
+get_directory_contents (char *path, dev_t device)
 {
   struct accumulator *accumulator;
 
@@ -176,8 +182,8 @@ get_directory_contents (char *path, int device)
     DIR *dirp = opendir (path);	/* for scanning directory */
     struct dirent *entry;	/* directory entry being scanned */
     char *name_buffer;		/* directory, `/', and directory member */
-    int name_buffer_size;	/* allocated size of name_buffer, minus 2 */
-    int name_length;		/* used length in name_buffer */
+    size_t name_buffer_size;	/* allocated size of name_buffer, minus 2 */
+    size_t name_length;		/* used length in name_buffer */
     struct directory *directory; /* for checking if already already seen */
     int all_children;
 
@@ -189,7 +195,7 @@ get_directory_contents (char *path, int device)
     errno = 0;			/* FIXME: errno should be read-only */
 
     name_buffer_size = strlen (path) + NAME_FIELD_SIZE;
-    name_buffer = xmalloc ((size_t) (name_buffer_size + 2));
+    name_buffer = xmalloc (name_buffer_size + 2);
     strcpy (name_buffer, path);
     if (path[strlen (path) - 1] != '/')
       strcat (name_buffer, "/");
@@ -202,103 +208,105 @@ get_directory_contents (char *path, int device)
 
     while (entry = readdir (dirp), entry)
       {
-	struct stat stat_data;
-
 	/* Skip `.' and `..'.  */
 
 	if (is_dot_or_dotdot (entry->d_name))
 	  continue;
 
-	if ((int) NAMLEN (entry) + name_length >= name_buffer_size)
-	  {
-	    while ((int) NAMLEN (entry) + name_length >= name_buffer_size)
-	      name_buffer_size += NAME_FIELD_SIZE;
-	    name_buffer = (char *)
-	      xrealloc (name_buffer, (size_t) (name_buffer_size + 2));
-	  }
-	strcpy (name_buffer + name_length, entry->d_name);
-
-	if (dereference_option
-#ifdef AIX
-	    ? statx (name_buffer, &stat_data, STATSIZE, STX_HIDDEN)
-	    : statx (name_buffer, &stat_data, STATSIZE, STX_HIDDEN | STX_LINK)
-#else
-	    ? stat (name_buffer, &stat_data)
-	    : lstat (name_buffer, &stat_data)
-#endif
-	    )
-	  {
-	    ERROR ((0, errno, _("Cannot stat %s"), name_buffer));
-	    continue;
-	  }
-
-	if ((one_file_system_option && device != stat_data.st_dev)
-	    || (exclude_option && check_exclude (name_buffer)))
+	if (excluded_filename (excluded, entry->d_name))
 	  add_to_accumulator (accumulator, "N", 1);
+	else
+	  {
+	    struct stat stat_data;
+
+	    if (NAMLEN (entry) + name_length >= name_buffer_size)
+	      {
+		while (NAMLEN (entry) + name_length >= name_buffer_size)
+		  name_buffer_size += NAME_FIELD_SIZE;
+		name_buffer = (char *)
+		  xrealloc (name_buffer, name_buffer_size + 2);
+	      }
+	    strcpy (name_buffer + name_length, entry->d_name);
+
+	    if (dereference_option
+#if STX_HIDDEN && !_LARGE_FILES /* AIX */
+		? statx (name_buffer, &stat_data, STATSIZE, STX_HIDDEN)
+		: statx (name_buffer, &stat_data, STATSIZE, STX_HIDDEN | STX_LINK)
+#else
+		? stat (name_buffer, &stat_data)
+		: lstat (name_buffer, &stat_data)
+#endif
+		)
+	      {
+		ERROR ((0, errno, _("Cannot stat %s"), name_buffer));
+		continue;
+	      }
+
+	    if (one_file_system_option && device != stat_data.st_dev)
+	      add_to_accumulator (accumulator, "N", 1);
 
 #ifdef AIX
-	else if (S_ISHIDDEN (stat_data.st_mode))
-	  {
-	    add_to_accumulator (accumulator, "D", 1);
-	    strcat (entry->d_name, "A");
-	    entry->d_namlen++;
-	  }
+	    else if (S_ISHIDDEN (stat_data.st_mode))
+	      {
+		add_to_accumulator (accumulator, "D", 1);
+		strcat (entry->d_name, "A");
+		entry->d_namlen++;
+	      }
 #endif
 
-	else if (S_ISDIR (stat_data.st_mode))
-	  {
-	    if (directory = find_directory (name_buffer), directory)
+	    else if (S_ISDIR (stat_data.st_mode))
 	      {
-		/* Devices having the high bit set are NFS devices, which are
-		   attributed somewhat randomly in automounting situations.
-		   For avoiding spurious incremental redumping of directories,
-		   we have to plainly consider all NFS devices as equal,
-		   relying on the i-node only to establish differences.  */
+		if (directory = find_directory (name_buffer), directory)
+		  {
+		    /* With NFS, the same file can have two different
+		       devices if an NFS directory is mounted in
+		       multiple locations, which is relatively common
+		       when automounting.  For avoiding spurious
+		       incremental redumping of directories, we have
+		       to plainly consider all NFS devices as equal,
+		       relying on the i-node only to establish
+		       differences.  */
 
-		/* FIXME: Gran Uddeborg <goeran@uddeborg.pp.se> says, on
-		   1996-09-20, that SunOS 5/Solaris 2 uses unsigned long for
-		   the device number type.  */
-
-		if ((((short) directory->device_number >= 0
-		      || (short) stat_data.st_dev >= 0)
-		     && directory->device_number != stat_data.st_dev)
-		    || directory->inode_number != stat_data.st_ino)
+		    if (! (((directory->nfs && NFS_FILE_STAT (stat_data))
+			    || directory->device_number == stat_data.st_dev)
+			   && directory->inode_number == stat_data.st_ino))
+		      {
+			if (verbose_option)
+			  WARN ((0, 0, _("Directory %s has been renamed"),
+				 name_buffer));
+			directory->allnew = 1;
+			directory->nfs = NFS_FILE_STAT (stat_data);
+			directory->device_number = stat_data.st_dev;
+			directory->inode_number = stat_data.st_ino;
+		      }
+		    directory->dir_text = "";
+		  }
+		else
 		  {
 		    if (verbose_option)
-		      WARN ((0, 0, _("Directory %s has been renamed"),
-			     name_buffer));
+		      WARN ((0, 0, _("Directory %s is new"), name_buffer));
+		    note_directory (name_buffer, &stat_data, "");
+		    directory = find_directory (name_buffer);
 		    directory->allnew = 1;
-		    directory->device_number = stat_data.st_dev;
-		    directory->inode_number = stat_data.st_ino;
 		  }
-		directory->dir_text = "";
-	      }
-	    else
-	      {
-		if (verbose_option)
-		  WARN ((0, 0, _("Directory %s is new"), name_buffer));
-		note_directory (name_buffer, stat_data.st_dev, stat_data.st_ino,
-				"");
-		directory = find_directory (name_buffer);
-		directory->allnew = 1;
-	      }
-	    if (all_children && directory)
-	      directory->allnew = 1;
+		if (all_children && directory)
+		  directory->allnew = 1;
 
-	    add_to_accumulator (accumulator, "D", 1);
+		add_to_accumulator (accumulator, "D", 1);
+	      }
+
+	    else
+	      if (!all_children
+		  && stat_data.st_mtime < newer_mtime_option
+		  && (!after_date_option
+		      || stat_data.st_ctime < newer_ctime_option))
+		add_to_accumulator (accumulator, "N", 1);
+	      else
+		add_to_accumulator (accumulator, "Y", 1);
 	  }
 
-	else
-	  if (!all_children
-	      && stat_data.st_mtime < newer_mtime_option
-	      && (!after_date_option
-		  || stat_data.st_ctime < newer_ctime_option))
-	    add_to_accumulator (accumulator, "N", 1);
-	  else
-	    add_to_accumulator (accumulator, "Y", 1);
-
 	add_to_accumulator (accumulator,
-			    entry->d_name, (int) (NAMLEN (entry) + 1));
+			    entry->d_name, NAMLEN (entry) + 1);
       }
     add_to_accumulator (accumulator, "\000\000", 2);
 
@@ -359,7 +367,7 @@ get_directory_contents (char *path, int device)
 `----------------------------------------------------------------------*/
 
 static void
-add_hierarchy_to_namelist (char *path, int device)
+add_hierarchy_to_namelist (char *path, dev_t device)
 {
   char *buffer = get_directory_contents (path, device);
 
@@ -375,14 +383,14 @@ add_hierarchy_to_namelist (char *path, int device)
 
   if (buffer)
     {
-      int name_length = strlen (path);
-      int allocated_length = (name_length >= NAME_FIELD_SIZE
-			      ? name_length + NAME_FIELD_SIZE
-			      : NAME_FIELD_SIZE);
-      char *name_buffer = xmalloc ((size_t) (allocated_length + 1));
+      size_t name_length = strlen (path);
+      size_t allocated_length = (name_length >= NAME_FIELD_SIZE
+				 ? name_length + NAME_FIELD_SIZE
+				 : NAME_FIELD_SIZE);
+      char *name_buffer = xmalloc (allocated_length + 1);
 				/* FIXME: + 2 above?  */
       char *string;
-      int string_length;
+      size_t string_length;
 
       strcpy (name_buffer, path);
       if (name_buffer[name_length - 1] != '/')
@@ -401,7 +409,7 @@ add_hierarchy_to_namelist (char *path, int device)
 		  while (name_length + string_length >= allocated_length)
 		    allocated_length += NAME_FIELD_SIZE;
 		  name_buffer = (char *)
-		    xrealloc (name_buffer, (size_t) (allocated_length + 1));
+		    xrealloc (name_buffer, allocated_length + 1);
 		}
 	      strcpy (name_buffer + name_length, string + 1);
 	      addname (name_buffer);
@@ -420,8 +428,6 @@ add_hierarchy_to_namelist (char *path, int device)
 static void
 read_directory_file (void)
 {
-  dev_t device_number;
-  ino_t inode_number;
   char *strp;
   FILE *fp;
   char buf[512];
@@ -473,24 +479,32 @@ read_directory_file (void)
 
   while (fgets (buf, sizeof (buf), fp))
     {
+      struct stat st;
+      long l;
+
       strp = &buf[strlen (buf)];
       if (strp[-1] == '\n')
 	strp[-1] = '\0';
       /* FIXME: For files ending with an incomplete line, maybe a NUL might
 	 be missing, here...  */
 
+      memset (&st, 0, sizeof st);
       strp = buf;
-      device_number = atol (strp);
+      st.st_dev = l = atol (strp);
+      if (st.st_dev != l)
+	ERROR ((0, 0, _("Device number out of range")));
       while (ISDIGIT (*strp))
 	strp++;
-      inode_number = atol (strp);
-      while (ISSPACE (*strp))
+      st.st_ino = l = atol (strp);
+      if (st.st_ino != l)
+	ERROR ((0, 0, _("Inode number out of range")));
+      while (ISSPACE ((unsigned char) *strp))
 	strp++;
       while (ISDIGIT (*strp))
 	strp++;
       strp++;
       unquote_string (strp);
-      note_directory (strp, device_number, inode_number, NULL);
+      note_directory (strp, &st, NULL);
     }
   if (fclose (fp) == EOF)
     ERROR ((0, errno, "%s", listed_incremental_option));
@@ -513,7 +527,7 @@ write_dir_file (void)
       ERROR ((0, errno, _("Cannot write to %s"), listed_incremental_option));
       return;
     }
-  fprintf (fp, "%lu\n", time_now);
+  fprintf (fp, "%lu\n", (unsigned long) time_now);
   for (directory = directory_list; directory; directory = directory->next)
     {
       if (!directory->dir_text)
@@ -521,13 +535,17 @@ write_dir_file (void)
       str = quote_copy_string (directory->name);
       if (str)
 	{
-	  fprintf (fp, "%u %u %s\n", directory->device_number,
-		   directory->inode_number, str);
+	  fprintf (fp, "%lu %lu %s\n",
+		   (unsigned long) directory->device_number,
+		   (unsigned long) directory->inode_number,
+		   str);
 	  free (str);
 	}
       else
-	fprintf (fp, "%u %u %s\n", directory->device_number,
-		 directory->inode_number, directory->name);
+	fprintf (fp, "%lu %lu %s\n",
+		 (unsigned long) directory->device_number,
+		 (unsigned long) directory->inode_number,
+		 directory->name);
     }
   if (fclose (fp) == EOF)
     ERROR ((0, errno, "%s", listed_incremental_option));
@@ -588,7 +606,7 @@ collect_and_sort_names (void)
 	  }
 
       if (
-#ifdef AIX
+#if STX_HIDDEN && !_LARGE_FILES /* AIX */
 	  statx (name->name, &statbuf, STATSIZE, STX_HIDDEN | STX_LINK)
 #else
 	  lstat (name->name, &statbuf) < 0
@@ -636,7 +654,8 @@ gnu_restore (int skipcrud)
   DIR *dirp;
   struct dirent *d;
   char *cur, *arc;
-  long size, copied;
+  off_t size;
+  size_t copied;
   union block *data_block;
   char *to;
 
@@ -649,7 +668,7 @@ gnu_restore (int skipcrud)
       /* The directory doesn't exist now.  It'll be created.  In any
 	 case, we don't have to delete any files out of it.  */
 
-      skip_file ((long) current_stat.st_size);
+      skip_file (current_stat.st_size);
       return;
     }
 
@@ -659,15 +678,18 @@ gnu_restore (int skipcrud)
       if (is_dot_or_dotdot (d->d_name))
 	continue;
 
-      add_to_accumulator (accumulator, d->d_name, (int) (NAMLEN (d) + 1));
+      add_to_accumulator (accumulator, d->d_name, NAMLEN (d) + 1);
     }
   closedir (dirp);
   add_to_accumulator (accumulator, "", 1);
 
   current_dir = get_accumulator (accumulator);
-  archive_dir = (char *) xmalloc ((size_t) current_stat.st_size);
+  size = current_stat.st_size;
+  if (size != current_stat.st_size)
+    FATAL_ERROR ((0, 0, _("Memory exhausted")));
+  archive_dir = (char *) xmalloc (size);
   to = archive_dir;
-  for (size = current_stat.st_size; size > 0; size -= copied)
+  for (; size > 0; size -= copied)
     {
       data_block = find_next_block ();
       if (!data_block)
@@ -678,7 +700,7 @@ gnu_restore (int skipcrud)
       copied = available_space_after (data_block);
       if (copied > size)
 	copied = size;
-      memcpy (to, data_block->buffer, (size_t) copied);
+      memcpy (to, data_block->buffer, copied);
       to += copied;
       set_next_block_after ((union block *)
 			    (data_block->buffer + copied - 1));
