@@ -1,5 +1,5 @@
 /* System dependent definitions for GNU tar.
-   Copyright (C) 1994, 1995, 1996, 1997 Free Software Foundation, Inc.
+   Copyright 1994, 1995, 1996, 1997, 1998, 1999 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -41,30 +41,27 @@ char *alloca ();
 #include <sys/types.h>
 
 /* Declare a generic pointer type.  */
-#if __STDC__ || defined(__TURBOC__)
+#if defined (__STDC__) || defined(__TURBOC__)
 # define voidstar void *
 #else
 # define voidstar char *
 #endif
 
-/* Declare ISASCII.  */
-
 #include <ctype.h>
 
+/* CTYPE_DOMAIN (C) is nonzero if the unsigned char C can safely be given
+   as an argument to <ctype.h> macros like `isspace'.  */
 #if STDC_HEADERS
-# define ISASCII(Char) 1
+# define CTYPE_DOMAIN(c) 1
 #else
-# ifdef isascii
-#  define ISASCII(Char) isascii (Char)
-# else
-#  if HAVE_ISASCII
-#   define ISASCII(Char) isascii (Char)
-#  else
-#   define ISASCII(Char) 1
-#  endif
-# endif
+# define CTYPE_DOMAIN(c) ((unsigned) (c) <= 0177)
 #endif
 
+#define ISDIGIT(c) ((unsigned) (c) - '0' <= 9)
+#define ISODIGIT(c) ((unsigned) (c) - '0' <= 7)
+#define ISPRINT(c) (CTYPE_DOMAIN (c) && isprint (c))
+#define ISSPACE(c) (CTYPE_DOMAIN (c) && isspace (c))
+
 /* Declare string and memory handling routines.  Take care that an ANSI
    string.h and pre-ANSI memory.h might conflict, and that memory.h and
    strings.h conflict on some systems.  */
@@ -167,9 +164,6 @@ extern int errno;
 # define S_ISVTX 0
 #endif
 
-#ifndef S_ISREG			/* POSIX.1 stat stuff missing */
-# define mode_t unsigned short
-#endif
 #if !defined(S_ISBLK) && defined(S_IFBLK)
 # define S_ISBLK(Mode) (((Mode) & S_IFMT) == S_IFBLK)
 #endif
@@ -206,9 +200,48 @@ extern int errno;
 #if !defined(S_ISCTG) && defined(S_IFCTG) /* contiguous file */
 # define S_ISCTG(Mode) (((Mode) & S_IFMT) == S_IFCTG)
 #endif
-#if !defined(S_ISVTX)
+#ifndef S_ISUID
+# define S_ISUID 0004000
+#endif
+#ifndef S_ISGID
+# define S_ISGID 0002000
+#endif
+#ifndef S_ISVTX
 # define S_ISVTX 0001000
 #endif
+#ifndef S_IRUSR
+# define S_IRUSR 0000400
+#endif
+#ifndef S_IWUSR
+# define S_IWUSR 0000200
+#endif
+#ifndef S_IXUSR
+# define S_IXUSR 0000100
+#endif
+#ifndef S_IRGRP
+# define S_IRGRP 0000040
+#endif
+#ifndef S_IWGRP
+# define S_IWGRP 0000020
+#endif
+#ifndef S_IXGRP
+# define S_IXGRP 0000010
+#endif
+#ifndef S_IROTH
+# define S_IROTH 0000004
+#endif
+#ifndef S_IWOTH
+# define S_IWOTH 0000002
+#endif
+#ifndef S_IXOTH
+# define S_IXOTH 0000001
+#endif
+
+#define MODE_WXUSR	(S_IWUSR | S_IXUSR)
+#define MODE_R		(S_IRUSR | S_IRGRP | S_IROTH)
+#define MODE_RW		(S_IWUSR | S_IWGRP | S_IWOTH | MODE_R)
+#define MODE_RWX	(S_IXUSR | S_IXGRP | S_IXOTH | MODE_RW)
+#define MODE_ALL	(S_ISUID | S_ISGID | S_ISVTX | MODE_RWX)
 
 #ifndef _POSIX_SOURCE
 # include <sys/param.h>
@@ -219,6 +252,26 @@ extern int errno;
 # include <unistd.h>
 #endif
 
+#ifndef SEEK_SET
+# define SEEK_SET 0
+#endif
+#ifndef SEEK_CUR
+# define SEEK_CUR 1
+#endif
+#ifndef SEEK_END
+# define SEEK_END 2
+#endif
+
+#ifndef STDIN_FILENO
+# define STDIN_FILENO 0
+#endif
+#ifndef STDOUT_FILENO
+# define STDOUT_FILENO 1
+#endif
+#ifndef STDERR_FILENO
+# define STDERR_FILENO 2
+#endif
+
 /* Declare make device, major and minor.  Since major is a function on
    SVR4, we have to resort to GOT_MAJOR instead of just testing if
    major is #define'd.  */
@@ -356,29 +409,33 @@ extern int errno;
 #endif
 
 /* Extract or fake data from a `struct stat'.  ST_NBLOCKS gives the
-   number of 512-byte blocks in the file (including indirect blocks).
-   fileblocks.c uses BSIZE.  HP-UX counts st_blocks in 1024-byte units,
+   number of ST_NBLOCKSIZE-byte blocks in the file (including indirect blocks).
+   HP-UX counts st_blocks in 1024-byte units,
    this loses when mixing HP-UX and BSD filesystems with NFS.  AIX PS/2
    counts st_blocks in 4K units.  */
 
 #if !HAVE_ST_BLOCKS
 # if defined(_POSIX_SOURCE) || !defined(BSIZE)
-#  define ST_NBLOCKS(Statbuf) (((Statbuf).st_size + 512 - 1) / 512)
+#  define ST_NBLOCKS(Statbuf) ((Statbuf).st_size / ST_NBLOCKSIZE + ((Statbuf).st_size % ST_NBLOCKSIZE != 0))
 # else
+   off_t st_blocks ();
 #  define ST_NBLOCKS(Statbuf) (st_blocks ((Statbuf).st_size))
 # endif
 #else
+# define ST_NBLOCKS(Statbuf) ((Statbuf).st_blocks)
 # if defined(hpux) || defined(__hpux__) || defined(__hpux)
-#  define ST_NBLOCKS(Statbuf) ((Statbuf).st_blocks * 2)
+#  define ST_NBLOCKSIZE 1024
 # else
 #  if defined(_AIX) && defined(_I386)
-#   define ST_NBLOCKS(Statbuf) ((Statbuf).st_blocks * 8)
-#  else
-#   define ST_NBLOCKS(Statbuf) ((Statbuf).st_blocks)
+#   define ST_NBLOCKSIZE (4 * 1024)
 #  endif
 # endif
 #endif
 
+#ifndef ST_NBLOCKSIZE
+#define ST_NBLOCKSIZE 512
+#endif
+
 /* This is a real challenge to properly get MTIO* symbols :-(.  ISC uses
    <sys/gentape.h>.  SCO and BSDi uses <sys/tape.h>; BSDi also requires
    <sys/tprintf.h> and <sys/device.h> for defining tp_dev and tpr_t.  It
@@ -443,6 +500,45 @@ off_t lseek ();
 # define DMALLOC_FUNC_CHECK
 # include <dmalloc.h>
 #endif
+
+#if HAVE_LIMITS_H
+# include <limits.h>
+#endif
+
+#ifndef CHAR_BIT
+# define CHAR_BIT 8
+#endif
+
+#ifndef CHAR_MAX
+# define CHAR_MAX TYPE_MAXIMUM (char)
+#endif
+
+#ifndef LONG_MAX
+# define LONG_MAX TYPE_MAXIMUM (long)
+#endif
+
+#if HAVE_INTTYPES_H
+# include <inttypes.h>
+#endif
+
+/* These macros work even on ones'-complement hosts (!).
+   The extra casts work around common compiler bugs.  */
+#define TYPE_SIGNED(t) (! ((t) 0 < (t) -1))
+#define TYPE_MINIMUM(t) (TYPE_SIGNED (t) \
+			 ? ~ (t) 0 << (sizeof (t) * CHAR_BIT - 1) \
+			 : (t) 0)
+#define TYPE_MAXIMUM(t) (~ (t) 0 - TYPE_MINIMUM (t))
+
+/* Bound on length of the string representing an integer value of type t.
+   Subtract one for the sign bit if t is signed;
+   302 / 1000 is log10 (2) rounded up;
+   add one for integer division truncation;
+   add one more for a minus sign if t is signed.  */
+#define INT_STRLEN_BOUND(t) \
+  ((sizeof (t) * CHAR_BIT - TYPE_SIGNED (t)) * 302 / 1000 \
+   + 1 + TYPE_SIGNED (t))
+
+#define UINTMAX_STRSIZE_BOUND (INT_STRLEN_BOUND (uintmax_t) + 1)
 
 /* Prototypes for external functions.  */
 
@@ -458,15 +554,17 @@ off_t lseek ();
 # include <locale.h>
 #endif
 #if !HAVE_SETLOCALE
-# define setlocale(Category, Locale)
+# define setlocale(Category, Locale) /* empty */
 #endif
 
 #if ENABLE_NLS
 # include <libintl.h>
 # define _(Text) gettext (Text)
 #else
-# define bindtextdomain(Domain, Directory)
-# define textdomain(Domain)
+# undef bindtextdomain
+# define bindtextdomain(Domain, Directory) /* empty */
+# undef textdomain
+# define textdomain(Domain) /* empty */
 # define _(Text) Text
 #endif
 #define N_(Text) Text
